@@ -136,7 +136,7 @@ import mimetypes
 rfc822_fmt = '%a, %d %b %Y %H:%M:%S GMT'
 iso8601_fmt = '%Y-%m-%dT%H:%M:%S.000Z'
 
-__version__ = "0.4"
+__version__ = "0.4.1"
 
 __all__ = "S3File", "S3Bucket", "S3Error"
 
@@ -145,6 +145,9 @@ def _amz_canonicalize(headers):
 
     >>> _amz_canonicalize({"x-amz-test": "test"})
     'x-amz-test:test\n'
+    >>> _amz_canonicalize({"x-amz-first": "test",
+    ...                    "x-amz-second": "hello"})
+    'x-amz-first:test\nx-amz-second:hello\n'
     >>> _amz_canonicalize({})
     ''
     """
@@ -153,7 +156,10 @@ def _amz_canonicalize(headers):
         header = header.lower()
         if header.startswith("x-amz-"):
             rv.setdefault(header, []).append(value)
-    return "".join(":".join((h, ",".join(v))) + "\n" for h, v in rv.iteritems())
+    parts = []
+    for key in sorted(rv):
+        parts.append("%s:%s\n" % (key, ",".join(rv[key])))
+    return "".join(parts)
 
 def metadata_headers(metadata):
     return dict(("X-AMZ-Meta-" + h, v) for h, v in metadata.iteritems())
@@ -194,12 +200,16 @@ def guess_mimetype(fn, default="application/octet-stream"):
     return mimetypes.guess_type(bfn + "." + ext)[0] or default
 
 def info_dict(headers):
-   return {"size": int(headers["content-length"]),
-           "mimetype": headers.get("content-type"),
-           "date": _rfc822_dt(headers["date"]),
-           "modify": _rfc822_dt(headers["last-modified"]),
-           "headers": headers,
-           "metadata": headers_metadata(headers)}
+    rv = {"headers": headers, "metadata": headers_metadata(headers)}
+    if "content-length" in headers:
+        rv["size"] = int(headers["content-length"])
+    if "content-type" in headers:
+        rv["mimetype"] = headers["content-type"]
+    if "date" in headers:
+        rv["date"] = _rfc822_dt(headers["date"]),
+    if "last-modified" in headers:
+        rv["modify"] = _rfc822_dt(headers["last-modified"])
+    return rv
 
 def name(o):
     """Find the name of *o*.
@@ -264,11 +274,14 @@ class S3Error(Exception):
     @classmethod
     def from_urllib(cls, e):
         """Try to read the real error from AWS."""
-        self = cls("HTTP error", code=e.code, url=e.filename)
-        self.code = e.code
-        self.fp = fp = e.fp
+        self = cls("HTTP error")
+        for attr in "reason", "code", "filename":
+            if hasattr(e, attr):
+                self.extra[attr] = getattr(e, attr)
+        fp = getattr(e, "fp", None)
         if not fp:
             return self
+        self.fp = fp
         # The latter part of this clause is to avoid some weird bug in urllib2
         # and AWS which has it read as if chunked, and AWS gives empty reply.
         try:
@@ -284,6 +297,9 @@ class S3Error(Exception):
                     self.msg += "..."
         return self
 
+    @property
+    def code(self): return self.extra.get("code")
+
 class StreamHTTPHandler(urllib2.HTTPHandler):
     pass
 
@@ -391,7 +407,7 @@ class S3Bucket(object):
         url = self.make_url(key, args)
         return AnyMethodRequest(method, url, data=data, headers=headers)
 
-    def make_url(self, key, args, arg_sep=";"):
+    def make_url(self, key, args=None, arg_sep=";"):
         url = self.base_url + "/"
         if key:
             url += aws_urlquote(key)
@@ -415,7 +431,10 @@ class S3Bucket(object):
                                        data=data, headers=headers)
             try:
                 return self.open_request(request)
-            except urllib2.HTTPError, e:
+            except (urllib2.HTTPError, urllib2.URLError), e:
+                # If S3 gives HTTP 500, we should try again.
+                if getattr(e, "code", None) == 500:
+                    continue
                 raise S3Error.from_urllib(e)
         else:
             raise RuntimeError("ran out of retries")  # Shouldn't happen.
@@ -496,11 +515,9 @@ class S3Bucket(object):
             if not data:
                 break
 
-    def url_for(self, key, authenticated=False, expires=None,
-            method='GET', content_md5='', mimetype='',
-            headers={}):
-        """
-        Produces the URL for given S3 object key.
+    def url_for(self, key, authenticated=False,
+                expire=datetime.timedelta(minutes=5)):
+        """Produces the URL for given S3 object key.
 
         *key* specifies the S3 object path relative to the
             base URL of the bucket.
@@ -508,38 +525,54 @@ class S3Bucket(object):
             to be used; such URLs include a signature and expiration
             time, and may be used by HTTP client apps to gain
             temporary access to private S3 objects.
-        *expires* indicates when the produced URL ceases to function,
-            in seconds from Jan 1, 1970 UTC; if omitted, the URL
+        *expire* indicates when the produced URL ceases to function,
+            in seconds from 1970-01-01T00:00:00 UTC; if omitted, the URL
             will expire in 5 minutes from now.
 
-        TODO: implement expires_in parameter that takes a
-            number of seconds from now or a positive timedelta object.
-
         URL for a publicly accessible S3 object:
-        >>> S3Bucket('bottle').url_for('the dregs')
+        >>> S3Bucket("bottle").url_for("the dregs")
         'https://s3.amazonaws.com/bottle/the%20dregs'
 
         Query string authenitcated URL example (fake S3 credentials are shown):
-        >>> b = S3Bucket('johnsmith', access_key='0PN5J17HBGZHT7JJ3X82',
-        ...     secret_key='uV3F3YluFJax1cknvbcGwgjvx4QpvB+leU8dUj2o')
-        >>> b.url_for('foo.js', authenticated=True) # doctest: +ELLIPSIS
+        >>> b = S3Bucket("johnsmith", access_key="0PN5J17HBGZHT7JJ3X82",
+        ...     secret_key="uV3F3YluFJax1cknvbcGwgjvx4QpvB+leU8dUj2o")
+        >>> b.url_for("foo.js", authenticated=True) # doctest: +ELLIPSIS
         'https://s3.amazonaws.com/johnsmith/foo.js?AWSAccessKeyId=...&Expires=...&Signature=...'
         """
-        args = tuple()
         if authenticated:
-            expires = time.time() + 5 * 60 if expires is None else expires
-            expires = str(long(expires))
-            auth_descriptor = ''.join((
-                method, '\n',
-                content_md5, '\n',
-                mimetype, '\n',
-                expires, '\n',
-                _amz_canonicalize(headers), # No '\n' by design!
-                self.canonicalized_resource(key) # No '\n' by design!
+            if not hasattr(expire, "timetuple"):
+                # XXX isinstance, but who uses UNIX timestamps?
+                if isinstance(expire, (int, long)):
+                    expire = datetime.datetime.fromtimestamp(expire)
+                else:
+                    # Assume timedelta.
+                    expire = datetime.datetime.now() + expire
+            expire_desc = str(long(time.mktime(expire.timetuple())))
+            auth_descriptor = "".join((
+                "GET\n",
+                "\n",
+                "\n",
+                expire_desc + "\n",
+                self.canonicalized_resource(key)  # No "\n" by design!
             ))
-            args = (
-                ('AWSAccessKeyId', self.access_key),
-                ('Expires', expires),
-                ('Signature', self.sign_description(auth_descriptor))
-            )
-        return self.make_url(key, args, '&')
+            args = (("AWSAccessKeyId", self.access_key),
+                    ("Expires", expire_desc),
+                    ("Signature", self.sign_description(auth_descriptor)))
+            return self.make_url(key, args, "&")
+        else:
+            return self.make_url(key)
+
+    def put_bucket(self, config_xml=None, acl=None):
+        if config_xml:
+            headers = {"Content-Length": len(config_xml),
+                       "Content-Type": "text/xml"}
+        else:
+            headers = {"Content-Length": "0"}
+        if acl:
+            headers["X-AMZ-ACL"] = acl
+        resp = self.make_request("PUT", key=None, data=config_xml, headers=headers)
+        resp.close()
+        return resp.code == 200
+
+    def delete_bucket(self):
+        return self.delete(None)
